# Built-in Types
const BoxIndex = UInt8 # Index of a box in the list of boxes
const TileIndex = Int32 # The index of a tile on the board - ie a position on the board
const TileValue = UInt32 # The contents of a board square
const ReachabilityCount = UInt32

# Tile flags
const BOX                     = one(TileValue)<<0
const FLOOR                   = one(TileValue)<<1
const GOAL                    = one(TileValue)<<2
const PLAYER                  = one(TileValue)<<3
const WALL                    = one(TileValue)<<4
const BOARD_PIECES = BOX + FLOOR + GOAL + PLAYER + WALL

const FLAG_BOX_REACHABLE_TILE = one(TileValue)<<5 # If true, this tile can reach a box start position
const FLAG_BOX_START_TILE     = one(TileValue)<<6
const FLAG_DOOR_TILE          = one(TileValue)<<7 # A floor with neighboring floors on both sides along one axis; all other neighbor squares are walls
const FLAG_GATE_TILE          = one(TileValue)<<8 # A floor with neighboring floors on both sides along at least one axis and splits the board in separate rooms
const FLAG_ILLEGAL_BOX_TILE   = one(TileValue)<<9

const FLAG_TUNNEL_UP          = one(TileValue)<<12
const FLAG_TUNNEL_LEFT        = one(TileValue)<<13
const FLAG_TUNNEL_DOWN        = one(TileValue)<<14
const FLAG_TUNNEL_RIGHT       = one(TileValue)<<15
const DIRECTION_TO_TUNNEL_FLAGS = FLAG_TUNNEL_UP + FLAG_TUNNEL_LEFT + FLAG_TUNNEL_DOWN + FLAG_TUNNEL_RIGHT

const FLAG_VISITED_TILE       = one(Int32)<<16 # Used to temporarily flag tiles when performing some calculations (ex: packing order)

# Each square on the board contains its goal number or packing order set number, if any, in the upper bits
const GOAL_BIT_SHIFT_COUNT = 24

# Move directions
# order must not change, e.g., the directions must be listed in alternate axis order
const DIR_UP = 0x01
const DIR_LEFT = 0x02
const DIR_DOWN = 0x03
const DIR_RIGHT = 0x04
const Direction = UInt8
const DIRECTIONS = DIR_UP:DIR_RIGHT
const N_DIRS = length(DIRECTIONS)
const NEXT_DIRECTION = [DIRECTIONS[mod1(dir+1,4)] for dir in DIRECTIONS]
const OPPOSITE_DIRECTION = [DIRECTIONS[mod1(dir+2,4)] for dir in DIRECTIONS]
const DIRECTION_TO_TUNNEL_FLAG = [FLAG_TUNNEL_UP, FLAG_TUNNEL_LEFT, FLAG_TUNNEL_DOWN, FLAG_TUNNEL_RIGHT]

# Board characters
const CH_BOX              = 'b'
const CH_BOX_ON_GOAL      = 'B'
const CH_GOAL             = '.'
const CH_FLOOR            = ' '
const CH_NON_BLANK_FLOOR  = '-'
const CH_PLAYER           = 'p'
const CH_PLAYER_ON_GOAL   = 'P'
const CH_SQUARE_SET       = '%'
const CH_WALL             = '#'

"""
A Board is a state represented as a list of tiles alongside the static game
information like floors / walls / goals.
The board tiles values can contain additional flags.
"""
const Board = Vector{TileValue} # Of length (Width+2)*(Height*2)